/*
 *  arch/ppc/kernel/head.S
 *
 * Derived from the below-mentioned work; 4xx series only
 * Copyright (C) 1998 Teleworld, Inc. All Rights Reserved.
 *
 * The 403 ignores the top address bit in real mode, and we load the kernel
 * at 0x800x0000, so there's no need to do any arithmetic on addresses. Additionally,
 * there are no segment registers or BAT registers (what were they there for in the
 * first place, especially in a RISC processor?) Instead, the TLB supports variable
 * page sizes. In real mode, the processor supports base/bounds style protection,
 * but we don't care about that. These guys seem to have learned a few lessons from
 * the MIPS architecture. -- jmb
 *
 *----------
 *
 *  PowerPC version 
 *    Copyright (C) 1995-1996 Gary Thomas (gdt@linuxppc.org)
 *
 *  Adapted for Power Macintosh by Paul Mackerras.
 *  Low-level exception handlers and MMU support
 *  rewritten by Paul Mackerras.
 *    Copyright (C) 1996 Paul Mackerras.
 *
 *  This file contains the low-level support and setup for the
 *  PowerPC platform, including trap and interrupt dispatch.
 *  Also included here is low-level thread/task switch support.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *	
 */

#include <linux/sys.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <asm/processor.h>
#include "ppc_asm.tmpl"
#include "ppc_defs.h"

#define SYNC() \
	sync; \
	isync

#define MSR_DEBUG	MSR_ME|MSR_IR|MSR_DR
#define MSR_CRITICAL	0
#define DBCR_SS		0x04000000

#define SRR2		0x3de
#define SRR3		0x3df

	.text
	.globl	_stext
_stext:
	.globl	_start
_start:
	lis	r0,KERNELBASE@h
	mtspr	0x3d6,r0		/* set up exception vector base */
	lis	r1,sys_stack@h		/* set up initial stack */
	ori	r1,r1,sys_stack@l
	li	r0,0
	stwu	r0,-STACK_FRAME_OVERHEAD(r1)
1:	lis	r7,start_here@ha
	addi	r7,r7,start_here@l
	mtlr	r7
	blr

/*
 * GCC sometimes accesses words at negative offsets from the stack
 * pointer, although the SysV ABI says it shouldn't.  To cope with
 * this, we leave this much untouched space on the stack on exception
 * entry.
 */
#define STACK_UNDERHEAD	64

/*
 * Macros for storing registers into and loading registers from
 * exception frames.
 */
#define SAVE_GPR(n, base)	stw	n,GPR0+4*(n)(base)
#define SAVE_2GPRS(n, base)	SAVE_GPR(n, base); SAVE_GPR(n+1, base)
#define SAVE_4GPRS(n, base)	SAVE_2GPRS(n, base); SAVE_2GPRS(n+2, base)
#define SAVE_8GPRS(n, base)	SAVE_4GPRS(n, base); SAVE_4GPRS(n+4, base)
#define SAVE_10GPRS(n, base)	SAVE_8GPRS(n, base); SAVE_2GPRS(n+8, base)
#define REST_GPR(n, base)	lwz	n,GPR0+4*(n)(base)
#define REST_2GPRS(n, base)	REST_GPR(n, base); REST_GPR(n+1, base)
#define REST_4GPRS(n, base)	REST_2GPRS(n, base); REST_2GPRS(n+2, base)
#define REST_8GPRS(n, base)	REST_4GPRS(n, base); REST_4GPRS(n+4, base)
#define REST_10GPRS(n, base)	REST_8GPRS(n, base); REST_2GPRS(n+8, base)

#define INCR(ctr,rega,regv)	\
	lis	rega,ctr@h;	\
	ori	rega,rega,ctr@l;\
	lwz	regv,0(rega);	\
	addi	regv,regv,1;	\
	stw	regv,0(rega)

/*
 * Exception entry code.  This code runs with address translation
 * turned off, i.e. using physical addresses.
 * We assume sprg3 has the physical address of the current
 * task's thread_struct.
 *
 */
#define EXCEPTION_PROLOG	\
0:	mtspr	SPRG0,r20;	\
	mtspr	SPRG1,r21;	\
	mfcr	r20;		\
	mfspr	r21,SRR1;		/* test whether from user or kernel */\
	andi.	r21,r21,MSR_PR;	\
	mr	r21,r1;			/* from kernel - use current sp */\
	beq	1f;		\
	mfspr	r21,SPRG3;		/* from user - load kernel sp */\
	lwz	r21,KSP(r21);	\
1: \
	subi	r21,r21,INT_FRAME_SIZE+STACK_UNDERHEAD;	/* alloc exc. frame */\
	stw	r1,GPR1(r21);	\
	stw	r1,0(r21);	\
	stw	r20,_CCR(r21);		/* save registers */		\
	stw	r22,GPR22(r21);	\
	stw	r23,GPR23(r21);	\
	mfspr	r20,SPRG0;	\
	stw	r20,GPR20(r21);	\
	mfspr	r22,SPRG1;	\
	stw	r22,GPR21(r21);	\
	mflr	r20;		\
	stw	r20,_LINK(r21);	\
	mfctr	r22;		\
	stw	r22,_CTR(r21);	\
	mfspr	r20,XER;	\
	stw	r20,_XER(r21);	\
	stw	r0,GPR0(r21);	\
	stw	r2,GPR2(r21);	\
	SAVE_4GPRS(3, r21);

#define CRITICAL_PROLOG		\
0:	mtspr	SPRG0,r20;	\
	mtspr	SPRG1,r21;	\
	mfcr	r20;		\
	mfspr	r21,SRR3;		/* test whether from user or kernel */\
	andi.	r21,r21,MSR_PR;	\
	mr	r21,r1;			/* from kernel - use current sp */\
	beq	1f;		\
	mfspr	r21,SPRG3;		/* from user - load kernel sp */\
	lwz	r21,KSP(r21);	\
1: \
	subi	r21,r21,INT_FRAME_SIZE+STACK_UNDERHEAD;	/* alloc exc. frame */\
	stw	r1,GPR1(r21);	\
	stw	r1,0(r21);	\
	stw	r20,_CCR(r21);		/* save registers */		\
	stw	r22,GPR22(r21);	\
	stw	r23,GPR23(r21);	\
	mfspr	r20,SPRG0;	\
	stw	r20,GPR20(r21);	\
	mfspr	r22,SPRG1;	\
	stw	r22,GPR21(r21);	\
	mflr	r20;		\
	stw	r20,_LINK(r21);	\
	mfctr	r22;		\
	stw	r22,_CTR(r21);	\
	mfspr	r20,XER;	\
	stw	r20,_XER(r21);	\
	stw	r0,GPR0(r21);	\
	stw	r2,GPR2(r21);	\
	SAVE_4GPRS(3, r21);

/*
 * Note: code which follows this uses cr0.eq (set if from kernel),
 * r21, r22 (SRR0), and r23 (SRR1).
 */

/*
 * Exception vectors.
 */
#define STD_EXCEPTION(n, label, hdlr)		\
	. = n;					\
label:						\
	EXCEPTION_PROLOG;			\
	mfspr	r22,SRR0;			\
	mfspr	r23,SRR1;			\
	addi	r3,r21,STACK_FRAME_OVERHEAD;	\
	li	r0,0;				\
	mr	r1,r21;				\
	lis	r20,MSR_KERNEL@h;		\
	ori	r20,r20,MSR_KERNEL@l;		\
	bl	transfer_to_handler; 		\
	.long	hdlr;				\
	.long	int_return

#define CRITICAL_EXCEPTION(n, label, hdlr)	\
	. = n;					\
label:						\
	CRITICAL_PROLOG;			\
	mfspr	r23,SPRG3;			\
	mfdcr	r22,0x90;	/* bus error address register */ \
	stw	r22,BEAR(r23);			\
	mfdcr	r22,0x91;	/* bus error syndrome register */ \
	stw	r22,BESR(r23);			\
	li	r22,0;				\
	mtdcr	0x91,r22;	/* must clear! */ \
	mfspr	r22,SRR2;			\
	mfspr	r23,SRR3;			\
	addi	r3,r21,STACK_FRAME_OVERHEAD;	\
	mr	r1,r21;				\
	li	r0,1;				\
	lis	r20,MSR_KERNEL@h;		\
	ori	r20,r20,MSR_KERNEL@l;		\
	bl	transfer_to_handler; 		\
	.long	hdlr;				\
	.long	critical_return

#define SYSCALL_EXCEPTION(n, label, hdlr)	\
	. = n;					\
label:						\
	EXCEPTION_PROLOG;			\
	mfspr	r22,SRR0;			\
	mfspr	r23,SRR1;			\
	addi	r3,r21,STACK_FRAME_OVERHEAD;	\
	li	r0,0;				\
	mr	r1,r21;				\
	lis	r20,MSR_KERNEL@h;		\
	ori	r20,r20,MSR_KERNEL@l;		\
	rlwimi  r20,r23,0,16,16;	/* copy EE bit from saved MSR */ \
	bl	transfer_to_handler; 		\
	.long	hdlr;				\
	.long	syscall_return

#ifdef CONFIG_XMON
#define CRITICAL_TRAP(n, label, msr)			\
	. = n;						\
label:							\
	mtspr	SPRG2,r21;				\
	lis	r21,debug_stack@ha;			\
	addi	r21,r21,debug_stack@l;			\
	subi	r21,r21,INT_FRAME_SIZE+STACK_UNDERHEAD;	\
	stw	r20,GPR20(r21);				\
	lis	r20,n@h;				\
	ori	r20,r20,n@l;				\
	stw	r20,TRAP(r21);				\
	mfspr	r20,SPRG2;				\
	stw	r20,GPR21(r21);				\
	stw	r7,GPR7(r21);				\
	lis	r7,msr@h;				\
	ori	r7,r7,msr@l

/* Critical interrupt pin */
	CRITICAL_TRAP(0x100, CriticalInterrupt, MSR_CRITICAL)
	b	FinishCritical
#else
	CRITICAL_EXCEPTION(0x100, CriticalInterrupt, MSR_CRITICAL)
#endif

/* Machine check */
	CRITICAL_EXCEPTION(0x200, MachineCheck, MachineCheckException)

/* Data access exception */
	STD_EXCEPTION(0x300, DataAccess, DataPageFault)

/* Instruction access exception */
	STD_EXCEPTION(0x400, InstructionAccess, InstructionPageFault)

/* External interrupt */
	STD_EXCEPTION(0x500, HardwareInterrupt, handle_IRQ)

/* Alignment exception */
	STD_EXCEPTION(0x600, Alignment, AlignmentException)

/* Program check exception */
	STD_EXCEPTION(0x700, ProgramCheck, ProgramCheckException)

	STD_EXCEPTION(0x800, Trap_08, UnknownException)
	STD_EXCEPTION(0x900, Trap_09, UnknownException)
	STD_EXCEPTION(0xa00, Trap_0a, UnknownException)
	STD_EXCEPTION(0xb00, Trap_0b, UnknownException)

/* System call */
/* Note: mis-implemented IMHO, since the strategy chosen
   for doing system calls requires a whole state save,
   which shouldn't be necessary */
	SYSCALL_EXCEPTION(0xc00, SystemCall, syscall);

	STD_EXCEPTION(0xd00, Trap_0d, UnknownException)
	STD_EXCEPTION(0xe00, Trap_0e, UnknownException)
	STD_EXCEPTION(0xf00, Trap_0f, UnknownException)

/* Programmable Interval Timer */
	STD_EXCEPTION(0x1000, Decrementer, timer_interrupt)

/* Inline TLB miss handling */

#define CRBIT	30
#define CRFIELD	7

/* Data TLB miss */
	. = 0x1100
DataTLBMiss:

	/* free up a few registers */
	mtspr	SPRG0,r20
	mtspr	SPRG1,r21
	mfspr	r21,SPRG3
#ifdef CONFIG_XMON
	lis	r20,DataTLBMiss@h
	ori	r20,r20,DataTLBMiss@l
	stw	r20,THREAD+TRAP(r21)
#endif
	stw	r22,THREAD+_TMP1(r21)		/* stash another register */
	stw	r23,THREAD+_TMP2(r21)		/* stash another register */
	mfcr	r20
	stw	r20,THREAD+_TLBCCR(r21)
	crset	CRBIT

	/* user or kernel mode? */
	mfmsr	r20
	andi.	r20,r20,MSR_PR
	bne	1f

	/* kernel mode, access to kernel address? */
	lis	r20,KERNELBASE@ha
	mfspr	r21,0x3d5
	cmplw	r21,r20
	ble	1f
	b	kernel_tlbmiss

/* Instruction TLB Miss */
/* Note: bigger than 64 instructions, but the following few
   exception vectors aren't used on the 403, so we just reuse
   that space here! */
	. = 0x1200
InstructionTLBMiss:

	/* free up a few registers */
	mtspr	SPRG0,r20
	mtspr	SPRG1,r21
	mfspr	r21,SPRG3
#ifdef CONFIG_XMON
	lis	r20,InstructionTLBMiss@h
	ori	r20,r20,InstructionTLBMiss@l
	stw	r20,THREAD+TRAP(r21)
#endif
	stw	r22,THREAD+_TMP1(r21)		/* stash another register */
	stw	r23,THREAD+_TMP2(r21)		/* stash another register */
	mfcr	r20
	stw	r20,THREAD+_TLBCCR(r21)
	crclr	CRBIT

	/* user or kernel mode? */
	mfmsr	r20
	andi.	r20,r20,MSR_PR
	bne	1f

	/* kernel mode, access to kernel address? */
	lis	r20,KERNELBASE@ha
	mfspr	r21,SRR0
	cmplw	r21,r20
	ble	1f

kernel_tlbmiss:
	/* set global SPID */
	li	r20,0
	mtspr	0x3b1,r20

	/* find kernel page table */
	.extern	swapper_pg_dir
	lis	r20,swapper_pg_dir@h
	ori	r20,r20,swapper_pg_dir@l
	b	2f
1:
	/* load user page table */
	mfspr	r20,SPRG3
	lwz	r20,THREAD+PGDIR(r20)
2:
	/* find second level page table */
	rlwinm	r22,r21,12,20,29
	lwzx	r20,r22,r20
	cmpwi	0,r20,0
	beq	longway

	/* find pte entry */
	rlwinm	r22,r21,22,20,29
	add	r20,r22,r20
	mfspr	r22,SPRG3
	stw	r20,THREAD+_PTE(r22)
	lwz	r20,0(r20)
	cmpwi	0,r20,0
	beq	longway

	/* page present? */
	andi.	r22,r20,0x1
	beq	longway

#ifdef CONFIG_XMON
	/*
	 * Damn, debugging mode. Save enough state to call the
	 * TLB checker.
	 * XXX the function is called in real mode!
	 */
	mfspr	r22,SRR1
	andi.	r22,r22,MSR_PR
	mr	r22,r1
	beq	25f

	mfspr	r22,SPRG3
	lwz	r22,KSP(r22)
25:
	subi	r22,r22,INT_FRAME_SIZE+STACK_UNDERHEAD
	stw	r1,GPR1(r22)
	mr	r1,r22
	stw	r2,GPR2(r1)
	mfspr	r2,SPRG3
	stw	r0,GPR0(r1)
	SAVE_10GPRS(3, r1)
	mflr	r22
	stw	r22,_LINK(r1)
	mfctr	r22
	stw	r22,_CTR(r1)
	mfxer	r22
	stw	r22,_XER(r1)
	mfspr	r22,SRR0
	stw	r22,_NIP(r1)
	mfspr	r22,SRR1
	stw	r22,_MSR(r1)
	lwz	r22,THREAD+TRAP(r2)
	andi.	r22,r22,0x3f00
	stw	r22,TRAP(r1)

	/*
	 * OK, compiler will save the rest.
	 */
	.globl	prohibit_if_guarded
	mr	r3,r21
	li	r4,4
	lis	r5,prohibit_if_guarded@h
	ori	r5,r5,prohibit_if_guarded@l
	mtctr	r5
	bctrl
	.globl	prohibit_return
prohibit_return:

	/*
	 * All done, put everything back.
	 */
	lwz	r22,_NIP(r1)
	mtspr	SRR0,r22
	lwz	r22,_MSR(r1)
	mtspr	SRR1,r22
	lwz	r22,_XER(r1)
	mtxer	r22
	lwz	r22,_CTR(r1)
	mtctr	r22
	lwz	r22,_LINK(r1)
	mtlr	r22
	REST_10GPRS(3, r1)
	lwz	r0,GPR0(r1)
	lwz	r2,GPR2(r1)
	lwz	r1,GPR1(r1)
#endif

	/* writing page? */
	bf	CRBIT,3f		/* no write checks for ITLBmiss */
	mfspr	r22,0x3d4
	andis.	r22,r22,0x80
	beq	3f

	/* is page writable? */
	andi.	r22,r20,0x4
	beq	longway

	/* mark it writable */
	ori	r20,r20,0x280		/* DIRTY|HWWRITE */
3:
	/* update pte */
	ori	r20,r20,0x100		/* ACCESSED */
	mfspr	r22,SPRG3
	lwz	r22,THREAD+_PTE(r22)
	stw	r20,0(r22)

	/* build low tlb entry */
	li	r22,0x200		/* TLBLO_EX */
	andi.	r23,r20,0x200		/* HWWRITE */
	beq	4f
	ori	r22,r22,0x100		/* TLBLO_WR */
4:
	andi.	r23,r20,0x20		/* NO_CACHE */
	beq	5f
	ori	r22,r22,0x4		/* TLBLO_I */
5:
	mfspr	r23,0x3b1		/* current spid ... */
	cmpwi	0,r23,0
	beq	6f
	ori	r22,r22,0x10		/* TLBLO_Z_USER */
6:
	rlwinm	r20,r20,0,0,19
	or	r20,r20,r22

	/* build high tlb entry */
	rlwinm	r22,r21,0,0,19
	ori	r22,r22,0xc0		/* VALID|4K */

	/* if already in TLB, drop a match over it */
	tlbsx.	r21,0,r22
	beq	8f
7:
	/* generate random location */
	mfspr	r21,0x3dd		/* time base - "random" */
	andi.	r21,r21,0x3f

	.extern	NTLB_WIRED
	lis	r23,NTLB_WIRED@ha
	lwz	r23,NTLB_WIRED@l(r23)
	cmpw	r23,r21
	blt	8f
	add	r21,r23,r21
8:
	/* drop in the final entry */
	tlbwe	r22,r21,0
	tlbwe	r20,r21,1
	sync

	/* everything worked, return to previous state */
	bt	CRBIT,9f
	INCR(_dtlbhit,r20,r21)
	b	10f
9:
	INCR(_itlbhit,r20,r21)
10:
	mfspr	r20,SPRG3
	lwz	r22,THREAD+_TMP1(r20)
	lwz	r23,THREAD+_TMP2(r20)
	lwz	r21,THREAD+_TLBCCR(r20)
	mtcr	r21
	lwz	r20,MM(r20)
	lwz	r20,MM_CONTEXT(r20)
	mtspr	0x3b1,r20
	mfspr	r21,SPRG1
	mfspr	r20,SPRG0
	rfi

longway:
	/* can't do it, let high level code deal with it */
	bf	CRBIT,11f
	INCR(_dtlbmiss,r20,r21)

	mfspr	r20,SPRG3
	lwz	r22,THREAD+_TLBCCR(r20)
	mtcr	r22
	lwz	r22,THREAD+_TMP1(r20)
	lwz	r23,THREAD+_TMP2(r20)
	mfspr	r21,SPRG1
	mfspr	r20,SPRG0
	b	RealDataTLBMiss
11:
	INCR(_itlbmiss,r20,r21)

	mfspr	r20,SPRG3
	lwz	r22,THREAD+_TLBCCR(r20)
	mtcr	r22
	lwz	r22,THREAD+_TMP1(r20)
	lwz	r23,THREAD+_TMP2(r20)
	mfspr	r21,SPRG1
	mfspr	r20,SPRG0
	b	RealInstructionTLBMiss

	STD_EXCEPTION(0x1600, RealInstructionTLBMiss, InstructionTLBMissException);
	STD_EXCEPTION(0x1700, RealDataTLBMiss, DataTLBMissException);

#if 0
/* don't waste the space */
	STD_EXCEPTION(0x1300, Trap_13, UnknownException)
	STD_EXCEPTION(0x1400, Trap_14, UnknownException)
	STD_EXCEPTION(0x1500, Trap_15, UnknownException)
	STD_EXCEPTION(0x1600, Trap_16, UnknownException)
	STD_EXCEPTION(0x1700, Trap_17, UnknownException)
	STD_EXCEPTION(0x1800, Trap_18, UnknownException)
	STD_EXCEPTION(0x1900, Trap_19, UnknownException)
	STD_EXCEPTION(0x1a00, Trap_1a, UnknownException)
	STD_EXCEPTION(0x1b00, Trap_1b, UnknownException)
	
	STD_EXCEPTION(0x1d00, Trap_1d, UnknownException)
	STD_EXCEPTION(0x1e00, Trap_1e, UnknownException)
	STD_EXCEPTION(0x1f00, Trap_1f, UnknownException)

#endif

/* Debugger and/or Critical exceptions */
#ifdef CONFIG_XMON
	CRITICAL_TRAP(0x2000, DebugTrap, MSR_DEBUG)

FinishCritical:
	mfcr	r20
	stw	r20,_CCR(r21)

	mfspr	r20,SRR3
	stw	r20,_MSR(r21)

	mfspr	r20,SRR2
	stw	r20,_NIP(r21)

	stw	r22,GPR22(r21)
	stw	r23,GPR23(r21)

	mflr	r20
	stw	r20,_LINK(r21)

	mfctr	r22
	stw	r22,_CTR(r21)

	mfspr	r20,XER
	stw	r20,_XER(r21)

	stw	r0,GPR0(r21)
	stw	r1,0(r21)
	stw	r1,GPR1(r21)
	stw	r2,GPR2(r21)
	SAVE_4GPRS(3, r21)
	SAVE_4GPRS(8, r21)
	SAVE_8GPRS(12, r21)
	SAVE_8GPRS(24, r21)

	addi	r3,r21,STACK_FRAME_OVERHEAD
	lis	r4,DebugException@h
	ori	r4,r4,DebugException@l
	mr	r1,r21

	lis	r22,_xdebug_exc@h
	ori	r22,r22,_xdebug_exc@l
	lis	r23,debug_return@h
	ori	r23,r23,debug_return@l

	mtspr	SRR2,r22
	mtspr	SRR3,r7
	mtlr	r23
	SYNC
	rfci

debug_return:
	lwz	r2,_CTR(r1)
	mtctr	r2

	lwz	r0,_LINK(r1)
	mtlr	r0

	lwz	r2,_XER(r1)
	mtspr	XER,r2

	lwz	r0,_CCR(r1)
	mtcrf	0xFF,r0

	REST_10GPRS(3, r1)
	REST_10GPRS(13, r1)
	REST_8GPRS(23, r1)
	REST_GPR(31, r1)

	lwz	r2,_NIP(r1)
	mtspr	SRR2,r2

	lwz	r0,_MSR(r1)
	mtspr	SRR3,r0

	lwz	r0,GPR0(r1)
	lwz	r2,GPR2(r1)
	lwz	r1,GPR1(r1)

	SYNC
	rfci
#else
	CRITICAL_EXCEPTION(0x2000, DebugTrap, DebugException)
#endif

/*
 * This code finishes saving the registers to the exception frame
 * and jumps to the appropriate handler for the exception, turning
 * on address translation.
 */
	.globl	transfer_to_handler
transfer_to_handler:
	stw	r22,_NIP(r21)
	lis	r22,MSR_WAIT@h
	andc	r23,r23,r22
	stw	r23,_MSR(r21)
	SAVE_GPR(7, r21)
	SAVE_4GPRS(8, r21)
	SAVE_8GPRS(12, r21)
	SAVE_8GPRS(24, r21)
	andi.	r23,r23,MSR_PR
	beq	2f
	mfspr	r23,SPRG3		/* if from user, fix up tss */
	lwz	r24,GPR1(r21)
	stw	r22,LAST_PC(r23)
	stw	r24,USER_STACK(r23)
	addi	r24,r1,STACK_FRAME_OVERHEAD
	stw	r24,PT_REGS(r23)
2:	mflr	r23
	andi.	r24,r23,0x3f00		/* get vector offset */
	stw	r24,TRAP(r21)
	li	r22,0
	stw	r22,RESULT(r21)
#ifdef CONFIG_XMON
	lis	r7,_xdebug@ha		/* check for breakout to debugger */
	addi	r7,r7,_xdebug@l
	lwz	r7,0(r7)
	andi.	r7,r7,0x2
	cmpi	0,r7,0
	bne	1f
	lwz	r24,0(r23)		/* virtual address of handler */
	b	2f
1:
	lwz	r4,0(r23)		/* actual handler */
	lwz	r5,4(r23)		/* where to go when done */
	lis	r24,_xdebug_exc@ha
	addi	r24,r24,_xdebug_exc@l
	lis	r20,MSR_DEBUG@h
	ori	r20,r20,MSR_DEBUG@l
	b	2f
#else
	lwz	r24,0(r23)		/* virtual address of handler */
#endif
2:
	lwz	r23,4(r23)		/* where to go when done */
	cmpwi	0,r0,0
	bne	3f

	mtspr	SRR0,r24		/* return from normal exception */
	mtspr	SRR1,r20
	mtlr	r23
	SYNC
	rfi				/* jump to handler, enable MMU */
3:
	mtspr	SRR2,r24		/* return from critical exception */
	mtspr	SRR3,r20
	mtlr	r23
	SYNC
	rfci				/* jump to handler, enable MMU */

#ifdef CONFIG_XMON
	.global call_realhandler
call_realhandler:
	lwz	r1,0(r1)		/* toss caller's stack frame */
	mtspr	SRR0,r4
	mtspr	SRR1,r5
	mtlr	r6
	SYNC
	rfi

/*
 * r3 - pointer to saved state
 * r4 - new stack pointer
 * r5 - MSR to set
 */
	.global kdebugtrap
kdebugtrap:
	lis	r0,DebugException@h
	ori	r0,r0,DebugException@l
	mtspr	SRR0,r0
	mtspr	SRR1,r5
	lis	r0,int_return@h
	ori	r0,r0,int_return@l
	mtlr	r0
	mr	r1,r4
	SYNC
	rfi
#endif

/*
 * This is where the main kernel code starts.
 * r3 should have the value 'TiVo' in it.  If so, r4 points to boot params.
 * MMU_init will put them someplace safe.
 */
start_here:
	/* Clear out the BSS */
	lis	r7,_end@ha
	addi	r7,r7,_end@l
	lis	r8,__bss_start@ha
	addi	r8,r8,__bss_start@l
	subf	r7,r8,r7
	addi	r7,r7,3
	rlwinm.	r7,r7,30,2,31
	beq	2f
	addi	r8,r8,-4
	mtctr	r7
	li	r0,0
3:	stwu	r0,4(r8)
	bdnz	3b
2:
	bl	MMU_init

/* Load up the kernel context */
2:	lis	r3,init_task@h
	ori	r3,r3,init_task@l
	addi	r3,r3,TSS	/* init task's TSS */
	mtspr	SPRG3,r3
	SYNC			/* Force all PTE updates to finish */

/* Now turn on the MMU for real! */
	lis	r4,MSR_KERNEL@h
	ori	r4,r4,MSR_KERNEL@l
	lis	r3,start_kernel@h
	ori	r3,r3,start_kernel@l
	mtspr	SRR0,r3
	mtspr	SRR1,r4
	rfi			/* enable MMU and jump to start_kernel */

/*
 * Switch between two kernel processes.
 */
_GLOBAL(_switch)
	/*
	 * Save current state
	 */
	stwu	r1,-INT_FRAME_SIZE-STACK_FRAME_OVERHEAD(r1)

        SAVE_10GPRS(14, r1)
        SAVE_8GPRS(24, r1)
        mflr    r0
	stw	r0,_LINK(r1)
	mfcr	r0
	stw	r0,_CCR(r1)
	stw     r1,KSP(r3)
	INCR(_switches, r14, r15);

	/*
	 * Load new state.
	 */
	mtspr   SPRG3,r4
	mtspr	0x3b1,r5
	lwz	r1,KSP(r4)
	lwz	r0,_LINK(r1)
	mtlr	r0
	lwz	r0,_CCR(r1)
	mtcr	r0
	REST_10GPRS(14,r1);
	REST_8GPRS(24, r1);

	addi	r1,r1,INT_FRAME_SIZE+STACK_FRAME_OVERHEAD
	blr

/*
 * Trap exit.
 */
	.globl	int_return
int_return:
0:	mfmsr	r30		/* Disable interrupts */
	li	r4,0
	ori	r4,r4,MSR_EE
	andc	r30,r30,r4
	SYNC			/* Some chip revs need this... */
	mtmsr	r30
	SYNC

	lis	r29,intr_count@ha /* Need to run 'bottom half' */
	lwz	r3,intr_count@l(r29)
	cmpi	0,r3,0
	bne	2f

	lwz	r5,_MSR(r1)
	and.	r5,r5,r4
	beq	2f

3:	lis	r4,lost_interrupts@ha
	lwz	r4,lost_interrupts@l(r4)
	cmpi	0,r4,0
	beq+	1f

	addi	r3,r1,STACK_FRAME_OVERHEAD
	bl	handle_IRQ
	b	3b

1:	lis	r4,bh_mask@ha
	lwz	r4,bh_mask@l(r4)
	lis	r5,bh_active@ha
	lwz	r5,bh_active@l(r5)
	and.	r4,r4,r5
	beq+	2f

	li	r3,1
	stw	r3,intr_count@l(r29)
	ori	r31,r30,MSR_EE	/* re-enable interrupts */
	SYNC
	mtmsr	r31
	SYNC

	bl	_EXTERN(do_bottom_half)
	SYNC
	mtmsr	r30		/* disable interrupts again */
	SYNC

	li	r0,0
	stw	r0,intr_count@l(r29)
2:	lwz	r3,_MSR(r1)	/* Returning to user mode? */
	andi.	r3,r3,MSR_PR
	beq+	10f		/* no - no need to mess with stack */

	lis	r31,current_set@ha
	lwz	r31,current_set@l(r31)
	lis	r3,need_resched@ha
	lwz	r3,need_resched@l(r3)
	cmpi	0,r3,0		/* check need_resched flag */
	beq+	7f

	bl	_EXTERN(schedule)
	b	0b

7:	lwz	r3,BLOCKED(r31)	/* Check for pending unblocked signals */
	lwz	r5,SIGNAL(r31)
	andc.	r0,r5,r3	/* Lets thru any unblocked */
	beq+	8f

	addi	r4,r1,STACK_FRAME_OVERHEAD
	bl	_EXTERN(do_signal)
	b	0b

8:	addi	r4,r1,INT_FRAME_SIZE+STACK_UNDERHEAD	/* size of frame */
	stw	r4,TSS+KSP(r31)	/* save kernel stack pointer */

10:	lwz	r2,_CTR(r1)
	lwz	r0,_LINK(r1)
	mtctr	r2
	mtlr	r0
	lwz	r2,_XER(r1)
	mtspr	XER,r2

#ifdef CONFIG_XMON
/* Force on the DE bit.  This should always be on if not in XMON */
	lwz	r2,_MSR(r1)
	lis	r3,MSR_DE@h
	ori	r3,r3,MSR_DE@l
	or	r2,r3,r2
	stw	r2,_MSR(r1)
#endif
	lis	r31,current_set@ha
	lwz	r31,current_set@l(r31)
	lwz	r3,TSS+TSS_FLAGS(r31)
	lis	r2,TASK_FLAG_SS@h
	ori	r2,r2,TASK_FLAG_SS@l
	and	r2,r2,r3
	cmplwi	r2,0
	beq	11f

	mfspr	r2,0x3f2
	lis	r3,DBCR_SS@h
	ori	r3,r3,DBCR_SS@l
	or	r2,r3,r2
	mtspr	0x3f2,r2

	lwz	r2,_MSR(r1)
	lis	r3,MSR_DE@h
	ori	r3,r3,MSR_DE@l
	or	r0,r3,r2
	b	12f
11:
	lwz	r0,_MSR(r1)
12:
	mtspr	SRR1,r0

	REST_10GPRS(3, r1)
	REST_10GPRS(13, r1)
	REST_8GPRS(23, r1)
	REST_GPR(31, r1)
	lwz	r2,_NIP(r1)	/* Restore environment */
	mtspr	SRR0,r2
	lwz	r0,_CCR(r1)
	mtcrf	0xFF,r0
	lwz	r0,GPR0(r1)
	lwz	r2,GPR2(r1)
	lwz	r1,GPR1(r1)
	SYNC
	rfi

	.global syscall_return
syscall_return:
	b	0b

/*
 * Return from critical exceptions.
 */
	.globl	critical_return
critical_return:
0:	mfmsr	r30		/* Disable interrupts */
	li	r4,0
	ori	r4,r4,MSR_EE
	andc	r30,r30,r4
	SYNC			/* Some chip revs need this... */
	mtmsr	r30
	SYNC
	lwz	r3,_MSR(r1)	/* Returning to user mode? */
	andi.	r3,r3,MSR_PR
	beq+	1f		/* no - no need to mess with stack */
	addi	r4,r1,INT_FRAME_SIZE+STACK_UNDERHEAD	/* size of frame */
	stw	r4,TSS+KSP(r31)	/* save kernel stack pointer */
1:	lwz	r2,_CTR(r1)
	lwz	r0,_LINK(r1)
	mtctr	r2
	mtlr	r0
	lwz	r2,_XER(r1)
	mtspr	XER,r2
	REST_10GPRS(3, r1)
	REST_10GPRS(13, r1)
	REST_8GPRS(23, r1)
	REST_GPR(31, r1)
	lwz	r2,_NIP(r1)	/* Restore environment */
	lwz	r0,_MSR(r1)
	mtspr	SRR2,r2
	mtspr	SRR3,r0
	lwz	r0,_CCR(r1)
	mtcrf	0xFF,r0
	lwz	r0,GPR0(r1)
	lwz	r2,GPR2(r1)
	lwz	r1,GPR1(r1)
	SYNC
	rfci

/*
 * Fake an interrupt from kernel mode.
 * This is used when enable_irq loses an interrupt.
 * We only fill in the stack frame minimally.
 */
_GLOBAL(fake_interrupt)
	mflr	r0
	stw	r0,4(r1)
	stwu	r1,-INT_FRAME_SIZE-STACK_UNDERHEAD(r1)
	stw	r0,_NIP(r1)
	stw	r0,_LINK(r1)
	mfmsr	r3
	stw	r3,_MSR(r1)
	li	r0,0x0fac
	stw	r0,TRAP(r1)
	addi	r3,r1,STACK_FRAME_OVERHEAD
	bl	handle_IRQ
	addi	r1,r1,INT_FRAME_SIZE+STACK_UNDERHEAD
	lwz	r0,4(r1)
	mtlr	r0
	blr

/*
 * This routine is just here to keep GCC happy - sigh...
 */	
_GLOBAL(__main)
	blr

/*
 * We put a few things here that have to be page-aligned.
 * This stuff goes at the beginning of the data segment,
 * which is page-aligned.
 */
	.data
	.globl	sdata
sdata:
	.space	2*4096
sys_stack:

#ifdef CONFIG_XMON
	.space	2*4096
debug_stack:
#endif

	.globl	empty_zero_page
empty_zero_page:
	.space	4096

	.globl	swapper_pg_dir
swapper_pg_dir:
	.space	4096	

/*
 * Miscellaneous statistics - gathered just for performance info
 */

	.globl	_head_stats
_head_stats:
_switches:
	.long	0
_itlbhit:
	.long	0
_dtlbhit:
	.long	0
_itlbmiss:
	.long	0
_dtlbmiss:
	.long	0
	.globl	_tlbpidwrap
_tlbpidwrap:
	.long	0
	
/*
 * The following must be in the same page.
 */
	.globl  xtime
xtime:
	.long   0
	.long   0

	.globl	Xppc_tblo
Xppc_tblo:
	.long	0

	.globl	Xcounts_per_usec
Xcounts_per_usec:
	.long	0

