/*************************************************************************
 *                                                                       *
 * Tivo TurboNet driver, Copyright (C) 2001,2002 Nick Kelsey.            *
 *   Web: www.silicondust.com                                            *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU General Public             *
 * License along with this library (see the file LICENSE.TXT); if not,   *
 * write to the Free Software Foundation, Inc., 59 Temple Place,         *
 * Suite 330, Boston, MA 02111-1307 USA.                                 *
 *                                                                       *
 *************************************************************************/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/if_ether.h>
#include <linux/ioport.h>
#include <linux/skbuff.h>
#include <linux/pci.h>
#include <linux/bios32.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mmu.h>
#include <asm/ppcops.h>

/*
 * ax88796 registers
 */
#define AX88796_REG_CR 0x00
#define AX88796_REG_DATA 0x10
#define AX88796_REG_IFGS1 0x12
#define AX88796_REG_IFGS2 0x13
#define AX88796_REG_MEMR 0x14
#define AX88796_REG_TR 0x15
#define AX88796_REG_IFG 0x16
#define AX88796_REG_GPI 0x17
#define AX88796_REG_GPOC 0x17
#define AX88796_REG_SPP_DPR 0x18
#define AX88796_REG_SPP_SPR 0x19
#define AX88796_REG_SPP_CPR 0x1A
#define AX88796_REG_RESET 0x1F

#define AX88796_PG0_PSTART 0x01
#define AX88796_PG0_PSTOP 0x02
#define AX88796_PG0_BNRY 0x03
#define AX88796_PG0_TSR 0x04
#define AX88796_PG0_TPSR 0x04
#define AX88796_PG0_NCR 0x05
#define AX88796_PG0_TBCR0 0x05
#define AX88796_PG0_CPR 0x06
#define AX88796_PG0_TBCR1 0x06
#define AX88796_PG0_ISR 0x07
#define AX88796_PG0_CRDA0 0x08
#define AX88796_PG0_RSAR0 0x08
#define AX88796_PG0_CRDA1 0x09
#define AX88796_PG0_RSAR1 0x09
#define AX88796_PG0_RBCR0 0x0A
#define AX88796_PG0_RBCR1 0x0B
#define AX88796_PG0_RSR 0x0C
#define AX88796_PG0_RCR 0x0C
#define AX88796_PG0_CNTR0 0x0D
#define AX88796_PG0_TCR 0x0D
#define AX88796_PG0_CNTR1 0x0E
#define AX88796_PG0_DCR 0x0E
#define AX88796_PG0_CNTR2 0x0F
#define AX88796_PG0_IMR 0x0F

#define AX88796_PG1_PAR0 0x01
#define AX88796_PG1_PAR1 0x02
#define AX88796_PG1_PAR2 0x03
#define AX88796_PG1_PAR3 0x04
#define AX88796_PG1_PAR4 0x05
#define AX88796_PG1_PAR5 0x06
#define AX88796_PG1_CPR 0x07
#define AX88796_PG1_MAR0 0x08
#define AX88796_PG1_MAR1 0x09
#define AX88796_PG1_MAR2 0x0A
#define AX88796_PG1_MAR3 0x0B
#define AX88796_PG1_MAR4 0x0C
#define AX88796_PG1_MAR5 0x0D
#define AX88796_PG1_MAR6 0x0E
#define AX88796_PG1_MAR7 0x0F

/*
 * ax88796 register fields
 */
#define AX88796_CR_PAGE0 (0 << 6)
#define AX88796_CR_PAGE1 (1 << 6)
#define AX88796_CR_DMA_OFF (4 << 3)
#define AX88796_CR_DMA_WR (2 << 3)
#define AX88796_CR_DMA_RD (1 << 3)
#define AX88796_CR_TXP (1 << 2)
#define AX88796_CR_START (1 << 1)
#define AX88796_CR_STOP (1 << 0)

#define AX88796_ISR_RST (1 << 7)
#define AX88796_ISR_RDC (1 << 6)
#define AX88796_ISR_CNT (1 << 5)
#define AX88796_ISR_OVW (1 << 4)
#define AX88796_ISR_TXE (1 << 3)
#define AX88796_ISR_RXE (1 << 2)
#define AX88796_ISR_PTX (1 << 1)
#define AX88796_ISR_PRX (1 << 0)

#define AX88796_IMR_RDCE (1 << 6)
#define AX88796_IMR_CNTE (1 << 5)
#define AX88796_IMR_OVWE (1 << 4)
#define AX88796_IMR_TXEE (1 << 3)
#define AX88796_IMR_RXEE (1 << 2)
#define AX88796_IMR_PTXE (1 << 1)
#define AX88796_IMR_PRXE (1 << 0)

#define AX88796_DCR_BOS (1 << 1)
#define AX88796_DCR_WTS (1 << 0)

#define AX88796_TCR_LOOP2 (2 << 1)
#define AX88796_TCR_LOOP1 (1 << 1)

#define AX88796_RCR_INTT (1 << 6)
#define AX88796_RCR_PRO (1 << 4)
#define AX88796_RCR_AM (1 << 3)
#define AX88796_RCR_AB (1 << 2)

/*
 * Packet buffers
 */
#define AX88796_TX_BUFF_COUNT 4
#define AX88796_TX_BUFF_START 0x4000
#define AX88796_TX_BUFF_STOP 0x5800
#define AX88796_RX_BUFF_START 0x5800
#define AX88796_RX_BUFF_STOP 0x8000

/*
 * Read/write configuration for tivo
 */
#define AX88796_BASE 0xF4000000
#define AX88796_OFFS_8(offs) ((offs << 3) | 0x06)
#define AX88796_OFFS_16(offs) ((offs << 3) | 0x02)
#define AX88796_IO_EXTENT 256
#define AX88796_IRQ 29

#define ax88796_in8(base, reg) __inb__((volatile u8 *)(base + AX88796_OFFS_8(reg)))
#define ax88796_in16(base, reg) __inw__((volatile u16 *)(base + AX88796_OFFS_16(reg)))
#define ax88796_out8(base, reg, v) __outb__(v, (volatile u8 *)(base + AX88796_OFFS_8(reg)))
#define ax88796_out16(base, reg, v) __outw__(v, (volatile u16 *)(base + AX88796_OFFS_16(reg)))

struct ax88796_local {
	struct enet_statistics stats;
	u8 rx_next;
	u8 tx_head;
	u8 tx_tail;
	u16 tx_length[AX88796_TX_BUFF_COUNT];
};

static u8 ax88796_name[8] = "ax88796";
static struct device ax88796_dev;

static char *macaddr = NULL;
MODULE_PARM(macaddr, "s");

/*
 * ax88796_dump_registers
 */
#if defined(DEBUG)
static void ax88796_dump_registers(void)
{
	u8 *base = (u8 *)AX88796_BASE;
	u8 cr;
	u8 p, r;
	
	printk("%s: ax88796_dump_registers...\n", ax88796_name);
	cr = ax88796_in8(base, 0);
	printk("%s: CR = 0x%02x\n", ax88796_name, cr);
	for (p = 0; p <= 1; p++) {
		ax88796_out8(base, 0, (cr & 0x3F) | (p << 6));
		for (r = 0; r < 31; r++) {
			printk("%s: p%1x r%02x = 0x%02x\n", ax88796_name, p, r, ax88796_in8(base, r));
		}		
	}
	ax88796_out8(base, 0, cr & 0x3F);
}
#endif

/*
 * hextol
 *     Trivial little hex-ASCII-to-integer decoder (0X prefix not allowed)
 */

static long int hextol(char *s)
{
   long sum = 0;
   char c;
   while (0 != (c = *s)) {
      if (c >= '0' && c <= '9') {
	 sum = (sum << 4) + (c - '0');
      } else if (c >= 'A' && c <= 'F') {
	 sum = (sum << 4) + (c - 'A' + 10);
      } else if (c >= 'a' && c <= 'f') {
	 sum = (sum << 4) + (c - 'a' + 10);
      } else {
	 return -1;
      }
      s++;
   }
   return sum;
}

/*
 * ax88796_reset
 *	Must be a function to make sure it doesn't get optimized out
 */
static u8 ax88796_reset(struct device *dev)
{
	u8 *base = (unsigned char *)dev->base_addr;
	return ax88796_in8(base, AX88796_REG_RESET);
}

/*
 * ax88796_set_tlb_entry
 */
static int ax88796_set_tlb_entry(void)
{
	int i;
	unsigned long tlbhi, new_tlbhi;
	unsigned long opid = mfspr(SPRN_PID);

	/* Find the existing TLB entry for this area and overwrite it */
	for (i = 0; i < 32; ++i) {
		asm("tlbre %0,%1,0" : "=r" (tlbhi) : "r" (i));
		if ((tlbhi & TLBHI_EPN) != IOBASE_8390)
			continue;
		/* found it! */
		new_tlbhi = (tlbhi & ~TLBHI_PGSZ) | TLBHI_PGSZ_1M;
		if (new_tlbhi != tlbhi) {
			asm("tlbwe %0,%1,0" : : "r" (new_tlbhi), "r" (i));
			asm("isync");
			printk("%s: updated tlb entry %d to 1M\n", ax88796_name, i);
		}
		mtspr(SPRN_PID, opid);
		return 0;
	}
	printk("%s: didn't find 8390 TLB entry!\n", ax88796_name);
	mtspr(SPRN_PID, opid);
	return -1;
}

/*
 * ax88796_send_packet
 */
static int ax88796_send_packet(struct sk_buff *skb, struct device *dev)
{
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	u8 *base = (u8 *)dev->base_addr;
	u16 *data;
	int length, count;

	/*
	 * Sanity checks
	 */
	if (dev->tbusy) {
		return 1;
	}
	if (skb == NULL) {
		return 0;
	}
	length = skb->len;
	if (skb->len <= 0) {
		return 0;
	}
	
	/*
	 * Disable interrupts
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);

	/*
	 * Upload packet
	 */
	count = (length + 7) & 0xFFF8;
	ax88796_out8(base, AX88796_PG0_RSAR0, 0x00);
	ax88796_out8(base, AX88796_PG0_RSAR1, (local->tx_head * 6) + (AX88796_TX_BUFF_START >> 8));
	ax88796_out8(base, AX88796_PG0_RBCR0, count);
	ax88796_out8(base, AX88796_PG0_RBCR1, count >> 8);

	count >>= 3;
	data = (u16 *)skb->data;
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_WR | AX88796_CR_START);
	while (count--) {
		ax88796_out16(base, AX88796_REG_DATA, *data++);
		ax88796_out16(base, AX88796_REG_DATA, *data++);
		ax88796_out16(base, AX88796_REG_DATA, *data++);
		ax88796_out16(base, AX88796_REG_DATA, *data++);
	}
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);

	/*
	 * Trigger send
	 */
	if (length < 60) {
		length = 60;
	}
	if (local->tx_head == local->tx_tail) {
		ax88796_out8(base, AX88796_PG0_TBCR0, length);
		ax88796_out8(base, AX88796_PG0_TBCR1, length >> 8);
		ax88796_out8(base, AX88796_PG0_TPSR, (local->tx_head * 6) + (AX88796_TX_BUFF_START >> 8));
		ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_TXP | AX88796_CR_START);
		local->tx_head = (local->tx_head + 1) & (AX88796_TX_BUFF_COUNT - 1);
	} else {
		local->tx_length[local->tx_head] = length;
		local->tx_head = (local->tx_head + 1) & (AX88796_TX_BUFF_COUNT - 1);
		if (local->tx_head == local->tx_tail) {
			dev->tbusy = 1;
		}
	}
	/*
	 * Enable interrupts
	 */
	ax88796_out8(base, AX88796_PG0_IMR, AX88796_IMR_PRXE | AX88796_IMR_PTXE | AX88796_IMR_TXEE);

	/*
	 * Complete
	 */
	dev_kfree_skb(skb, FREE_WRITE);
	return 0;
}

/*
 * ax88796_get_packet
 */
static int ax88796_get_packet(struct device *dev)
{
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	struct sk_buff *skb;
	u8 *base = (u8 *)dev->base_addr;
	u16 *data;
	u8 curr, bnry, next;
	u16 length, count;

	/*
	 * Any packet to read out?
	 */
	curr = ax88796_in8(base, AX88796_PG0_CPR);
	if (curr == local->rx_next) {
		return -1;
	}

	/*
	 * Read out the packet information header (4 bytes)
	 */
	ax88796_out8(base, AX88796_PG0_RSAR0, 0);
	ax88796_out8(base, AX88796_PG0_RSAR1, local->rx_next);
	ax88796_out8(base, AX88796_PG0_RBCR0, 4);
	ax88796_out8(base, AX88796_PG0_RBCR1, 0);

	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_RD | AX88796_CR_START);
	next = ax88796_in16(base, AX88796_REG_DATA) & 0x00FF;
	length = ax88796_in16(base, AX88796_REG_DATA);
	length = (length << 8) | (length >> 8);
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);

	/*
	 * Sanity check the packet size
	 */
	if ((length < 64) || (length > 1522)) {
		/* Purge buffers */
		local->rx_next = curr;
		bnry = curr - 1;
		if (bnry < (AX88796_RX_BUFF_START >> 8)) {
			bnry = (AX88796_RX_BUFF_STOP >> 8) - 1;
		}
		ax88796_out8(base, AX88796_PG0_BNRY, bnry);
		#if defined(DEBUG)
		printk("%s: rx invalid length 0x%04x\n", ax88796_name, length);
		#endif
		return -1;
	}
	
	/*
	 * Caculate new bnry pointer for when xfer is complete
	 */
	bnry = next - 1;
	if (bnry < (AX88796_RX_BUFF_START >> 8)) {
		bnry = (AX88796_RX_BUFF_STOP >> 8) - 1;
	}
	
	/*
	 * Create packet
	 */
	count = (length + 7) & 0xFFF8;
	skb = dev_alloc_skb(count + 2);
	if (skb == NULL) {
		/* Drop packet */
		ax88796_out8(base, AX88796_PG0_BNRY, bnry);
		local->rx_next = next;
		#if defined(DEBUG)
		printk("%s: rx out of memory\n", ax88796_name);
		#endif
	} else {
		skb_reserve(skb, 2);	/* Initial offset of 2 so that the ethernet header ends on a 32-bit boundary */
		skb->dev = dev;
		data = (u16 *)skb_put(skb, length);
		
		ax88796_out8(base, AX88796_PG0_RSAR0, 4);
		ax88796_out8(base, AX88796_PG0_RSAR1, local->rx_next);
		ax88796_out8(base, AX88796_PG0_RBCR0, count);
		ax88796_out8(base, AX88796_PG0_RBCR1, count >> 8);

		count >>= 3;		
		ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_RD | AX88796_CR_START);
		while (count--) {
			*data++ = ax88796_in16(base, AX88796_REG_DATA);
			*data++ = ax88796_in16(base, AX88796_REG_DATA);
			*data++ = ax88796_in16(base, AX88796_REG_DATA);
			*data++ = ax88796_in16(base, AX88796_REG_DATA);
		}
		ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);
		
		ax88796_out8(base, AX88796_PG0_BNRY, bnry);
		local->rx_next = next;

		skb->protocol = eth_type_trans(skb, dev);
		netif_rx(skb);
	}

	/*
	 * Complete (success)
	 */
	return 0;
}

/*
 * ax88796_overflow
 */
static void ax88796_overflow(struct device *dev)
{
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	u8 *base = (u8 *)dev->base_addr;
	u8 resend;

	#if defined(DEBUG)
	printk("%s: overflow!\n", ax88796_name);
	#endif
	
	/*
	 * 1: Read and store the value of TXP.
	 */
	resend = ax88796_in8(base, AX88796_REG_CR) & AX88796_CR_TXP;

	/*
	 * 2: Issue the STOP command.
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_STOP);

	/*
	 * 3: Wait for at least 1.6 ms.
	 */
	udelay(2000);

	/*
	 * 4: Clear the NIC's remote byte count registers.
	 */
	ax88796_out8(base, AX88796_PG0_RBCR0, 0);
	ax88796_out8(base, AX88796_PG0_RBCR1, 0);

	/*
	 * 5: Read the stored value of TXP from step 1.  If this value is a 1
 	 *    read the NIC's interrupt status register.  If either the PTX or TXE
 	 *    bits are set, set the "resend" variable to 0.
	 */
 	if (resend) {
	 	if (ax88796_in8(base, AX88796_PG0_ISR) & (AX88796_ISR_PTX | AX88796_ISR_TXE)) {
	 		resend = 0;
	 	}
	 }

	/*
	 * 6: Place the NIC in loopback mode 1.
	 */
	ax88796_out8(base, AX88796_PG0_TCR, AX88796_TCR_LOOP1);

	/*
	 * 7: Issue the START command.
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);

	/*
	 * 8: Read out packets.
	 */
	while (ax88796_get_packet(dev) == 0) {
		local->stats.rx_packets++;
	}

	/*
	 * 9: Reset the OVW bit in the interrupt status register.
	 */
	ax88796_out8(base, AX88796_PG0_ISR, AX88796_ISR_OVW);

	/*
	 * 10: Take the NIC out of loopback.
	 */
	ax88796_out8(base, AX88796_PG0_TCR, 0x00);

	/*
	 * 11: If the resend variable is set then reissue the transmit command.
	 */
	if (resend) {
		ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_TXP | AX88796_CR_START);
	}
}

/*
 * ax88796_interrupt
 */
static void ax88796_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct device *dev = &ax88796_dev;
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	u8 *base = (u8 *)dev->base_addr;
	u8 isr;

	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);
	dev->interrupt = 1;

	/*
	 * Read ISR flags
	 */
	isr = 0;
	while(1) {
		isr |= ax88796_in8(base, AX88796_PG0_ISR);
		if ((isr & (AX88796_ISR_OVW | AX88796_ISR_PTX | AX88796_ISR_TXE | AX88796_ISR_PRX)) == 0) {
			break;
		}

		/*
		 * Check for overflow
		 */
		if (isr & AX88796_ISR_OVW) {
			ax88796_overflow(dev);
			isr &= ~AX88796_ISR_OVW;
		}
	
		/*
		 * Check for tx complete
		 */
		if (isr & (AX88796_ISR_PTX | AX88796_ISR_TXE)) {
			ax88796_out8(base, AX88796_PG0_ISR, AX88796_ISR_PTX | AX88796_ISR_TXE);
			isr &= ~(AX88796_ISR_PTX | AX88796_ISR_TXE);
			local->tx_tail = (local->tx_tail + 1) & (AX88796_TX_BUFF_COUNT - 1);
			if (local->tx_head != local->tx_tail) {
				ax88796_out8(base, AX88796_PG0_TBCR0, local->tx_length[local->tx_tail]);
				ax88796_out8(base, AX88796_PG0_TBCR1, local->tx_length[local->tx_tail] >> 8);
				ax88796_out8(base, AX88796_PG0_TPSR, (local->tx_tail * 6) + (AX88796_TX_BUFF_START >> 8));
				ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_TXP | AX88796_CR_START);
			}
			dev->tbusy = 0;
			local->stats.tx_packets++;
		}
	
		/*
		 * Check for rx packets
		 */
		if (isr & AX88796_ISR_PRX) {
			ax88796_out8(base, AX88796_PG0_ISR, AX88796_ISR_PRX);
			if (ax88796_get_packet(dev) == 0) {
				local->stats.rx_packets++;
			} else {
				isr &= ~AX88796_ISR_PRX;
			}
		}
	}

	dev->interrupt = 0;
}

/*
 * ax88796_get_stats
 */
static struct enet_statistics *ax88796_get_stats(struct device *dev)
{
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	return &local->stats;
}

/*
 * ax88796_set_multicast_list
 */
static void ax88796_set_multicast_list(struct device *dev)
{
}

/*
 * ax88796_change_mtu
 */
static int ax88796_change_mtu(struct device *dev, int mtu)
{
	if (mtu > 1500) {
		return -EINVAL;
	}

	dev->mtu = mtu;
	return 0;
}

/*
 * ax88796_open
 */
static int ax88796_open(struct device *dev)
{
	struct ax88796_local *local = (struct ax88796_local *)dev->priv;
	u8 *base = (u8 *)dev->base_addr;

	/*
	 * Reset AX88796
	 */
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);
	ax88796_reset(dev);
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);
	udelay(2000);

	/*
	 * Select device stopped, no DMA and register page 0
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_STOP);

	/*
	 * Set word-wide access
	 */
	ax88796_out8(base, AX88796_PG0_DCR, AX88796_DCR_WTS);

	/*
	 * Clear down the remote byte count registers.
	 */
	ax88796_out8(base, AX88796_PG0_RBCR0, 0);
	ax88796_out8(base, AX88796_PG0_RBCR1, 0);

	/*
	 * Configure receive - accept broadcast, multicast, and matching MAC.
	 */
	ax88796_out8(base, AX88796_PG0_RCR, AX88796_RCR_INTT | AX88796_RCR_AB | AX88796_RCR_AM);

	/*
	 * Configure transmit - loopback.
	 */
	ax88796_out8(base, AX88796_PG0_TCR, AX88796_TCR_LOOP1);

	/*
	 * Configure receive buffer.
	 */
	ax88796_out8(base, AX88796_PG0_PSTART, AX88796_RX_BUFF_START >> 8);
	ax88796_out8(base, AX88796_PG0_PSTOP, AX88796_RX_BUFF_STOP >> 8);

	/*
	 * Reset boundary pointer.
	 */
	ax88796_out8(base, AX88796_PG0_BNRY, AX88796_RX_BUFF_START >> 8);

	/*
	 * Clear the ISR flags and enable transmit and receive interrupts.
	 */
	ax88796_out8(base, AX88796_PG0_ISR, 0xff);
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);

	/*
	 * Select device stopped, no DMA and register page 1.
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE1 | AX88796_CR_DMA_OFF | AX88796_CR_STOP);

	/*
	 * Set MAC address
	 */
	ax88796_out8(base, AX88796_PG1_PAR0, dev->dev_addr[0]);
	ax88796_out8(base, AX88796_PG1_PAR1, dev->dev_addr[1]);
	ax88796_out8(base, AX88796_PG1_PAR2, dev->dev_addr[2]);
	ax88796_out8(base, AX88796_PG1_PAR3, dev->dev_addr[3]);
	ax88796_out8(base, AX88796_PG1_PAR4, dev->dev_addr[4]);
	ax88796_out8(base, AX88796_PG1_PAR5, dev->dev_addr[5]);

	/*
	 * Clear multicast table (accept all)
	 */
	ax88796_out8(base, AX88796_PG1_MAR0, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR1, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR2, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR3, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR4, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR5, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR6, 0xff);
	ax88796_out8(base, AX88796_PG1_MAR7, 0xff);

	/*
	 * Reset Current pointer.
	 */
	local->rx_next = (AX88796_RX_BUFF_START >> 8) + 1;
	ax88796_out8(base, AX88796_PG1_CPR, local->rx_next);

	/*
	 * Enter start mode (still in loopback), page 0
	 */
	ax88796_out8(base, AX88796_REG_CR, AX88796_CR_PAGE0 | AX88796_CR_DMA_OFF | AX88796_CR_START);

	/*
	 * Configure transmit and start normal operation.
	 */
	ax88796_out8(base, AX88796_PG0_TCR, 0);

	/*
	 * Complete
	 */
	local->tx_head = 0;
	local->tx_tail = 0;
	dev->tbusy = 0;
	dev->start = 1;
	
	/*
	 * Enable interrupts
	 */
	ax88796_out8(base, AX88796_PG0_IMR, AX88796_IMR_PRXE | AX88796_IMR_PTXE | AX88796_IMR_TXEE);

	return 0;
}

/*
 * ax88796_close
 */
static int ax88796_close(struct device *dev)
{
	u8 *base = (u8 *)dev->base_addr;

	/*
	 * Reset AX88796
	 */
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);
	ax88796_reset(dev);
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);
	udelay(2000);

	/*
	 * Complete
	 */
	dev->start = 0;
	return 0;
}

/*
 * ax88796_init
 */
static int ax88796_init(struct device *dev)
{
	u8 *base = (u8 *)dev->base_addr;
	u32 iocr;
	u8 reg;

	/*
	 * Make sure resources are not already in use
	 */
	if (check_region(dev->base_addr, AX88796_IO_EXTENT) != 0) {
		printk("%s: IO region already in use at 0x%lx\n", ax88796_name, dev->base_addr);
		return -EAGAIN;
	}

	/*
	 * Configure PPC bus timing & interrupt
	 */
	if (ax88796_set_tlb_entry() != 0) {
		return -ENODEV;
	}
	asm volatile("sync");
	mtdcr(SPRN_BR5, 0x40194663);
	iocr = mfdcr(SPRN_IOCR);
	mtdcr(SPRN_IOCR, (iocr & ~(0x0C000000)) | 0x04000000);
	asm volatile("isync");

	/*
	 * Check for AX88796
	 */
	ax88796_reset(dev);
	ax88796_out8(base, AX88796_PG0_IMR, 0x00);
	udelay(2000);
	reg = ax88796_in8(base, AX88796_PG0_ISR);
	if ((reg & AX88796_ISR_RST) == 0) {
		#if defined(DEBUG)
		printk("%s: No reset response (isr = 0x%02x)\n", ax88796_name, reg);
		#endif
		printk("%s: NIC not found at 0x%lx\n", ax88796_name, dev->base_addr);
		return -ENODEV;
	}
	reg = ax88796_in8(base, AX88796_REG_IFGS1);
	if (reg != 0x0C) {
		#if defined(DEBUG)
		printk("%s: IFCS1 check failed (ifgs1 = 0x%02x)\n", ax88796_name, reg);
		#endif
		printk("%s: NIC not found at 0x%lx\n", ax88796_name, dev->base_addr);
		return -ENODEV;
	}
	reg = ax88796_in8(base, AX88796_REG_IFGS2);
	if (reg != 0x12) {
		#if defined(DEBUG)
		printk("%s: IFCS2 check failed (ifgs2 = 0x%02x)\n", ax88796_name, reg);
		#endif
		printk("%s: NIC not found at 0x%lx\n", ax88796_name, dev->base_addr);
		return -ENODEV;
	}

	ax88796_out8(base, AX88796_PG0_IMR, 0);

	/*
	 * Allocate memory
	 */
	dev->priv = kmalloc(sizeof(struct ax88796_local), GFP_KERNEL);
	if (dev->priv == NULL) {
		printk ("%s: kmalloc failed\n", ax88796_name);
		return -ENOMEM;
	}
	memset(dev->priv, 0, sizeof(struct ax88796_local));

	/*
	 * Init resources
	 */
	request_region(dev->base_addr, AX88796_IO_EXTENT, ax88796_name);
	if (request_irq(dev->irq, &ax88796_interrupt, 0, ax88796_name, NULL) != 0) {
		release_region(dev->base_addr, AX88796_IO_EXTENT);
		printk ("%s: request_irq failed\n", ax88796_name);
		return -EAGAIN;
	}

	/*
	 * Init dev structure
	 */
	dev->dev_addr[0] = 0x00;
	dev->dev_addr[1] = 0x0B;
	dev->dev_addr[2] = 0xAD;
	dev->dev_addr[3] = 0x00;
	dev->dev_addr[4] = 0xAD;
	dev->dev_addr[5] = 0xD0;
        // The following is an ugly hack... dplatt
        if (macaddr && strlen(macaddr) >= 6) {
	   long digits = hextol(macaddr + strlen(macaddr) - 6);
	   dev->dev_addr[5] = digits & 0xFF;
	   dev->dev_addr[4] = (digits >> 8) & 0xFF;
	   dev->dev_addr[3] = (digits >> 16) & 0xFF;
	   printk("%s: using MAC address %02X:%02X:%02X:%02X:%02X:%02X\n",
		  ax88796_name,
		  dev->dev_addr[0], dev->dev_addr[1], dev->dev_addr[2],
		  dev->dev_addr[3], dev->dev_addr[4], dev->dev_addr[5]);
	}
	dev->mtu = 1500;
	dev->open = &ax88796_open;
    	dev->stop = &ax88796_close;
	dev->hard_start_xmit = &ax88796_send_packet;
	dev->get_stats = &ax88796_get_stats;
	dev->set_multicast_list = &ax88796_set_multicast_list;
	dev->change_mtu = &ax88796_change_mtu;
	ether_setup(dev);

	/*
	 * Success
	 */
	return 0;
}

/*
 * module interface
 */
static char devicename[9] = "eth0";
static int io = AX88796_BASE;
static int irq = AX88796_IRQ;

/*
 * init_module
 */
int init_module(void)
{
	struct device *dev = &ax88796_dev;
	int result;
	
        if (check_region(0xFFB80000, 8)) {
	   /* ugh - dplatt */
           printk("%s: skipped, debug card present!\n", ax88796_name);
           return -ENODEV;
        }

        /*
	 * Fill in network device structure
	 */
	memset(dev, 0, sizeof(struct device));
	dev->name = devicename;
	dev->base_addr = io;
	dev->irq = irq;
	dev->mem_end = 0;
	dev->init = &ax88796_init;

	/*
	 * Register network device
	 */
	result = 0;
	result = register_netdev(dev);
	if (result != 0) {
		printk("%s: driver installation failed\n", ax88796_name);
		return result;
	}

	/*
	 * Complete
	 */
	printk("%s: driver installed\n", ax88796_name);
	return 0;
}

/*
 * cleanup_module
 */
void cleanup_module(void)
{
	struct device *dev = &ax88796_dev;

	unregister_netdev(dev);
	free_irq(dev->irq, NULL);
	irq2dev_map[dev->irq] = NULL;
	release_region(dev->base_addr, AX88796_IO_EXTENT);
	if (dev->priv) {
		kfree_s(dev->priv, sizeof(struct ax88796_local));
	}
}
