/*
 * PowerPC memory management structures
 */

#ifndef _PPC_MMU_H_
#define _PPC_MMU_H_

#if 0
/* Hardware Page Table Entry */
typedef struct _PTE
   {
      unsigned long v:1;	/* Entry is valid */
      unsigned long vsid:24;	/* Virtual segment identifier */
      unsigned long h:1;	/* Hash algorithm indicator */
      unsigned long api:6;	/* Abbreviated page index */
      unsigned long rpn:20;	/* Real (physical) page number */
      unsigned long    :3;	/* Unused */
      unsigned long r:1;	/* Referenced */
      unsigned long c:1;	/* Changed */
      unsigned long w:1;	/* Write-thru cache mode */
      unsigned long i:1;	/* Cache inhibited */
      unsigned long m:1;	/* Memory coherence */
      unsigned long g:1;	/* Guarded */
      unsigned long    :1;	/* Unused */
      unsigned long pp:2;	/* Page protection */
   } PTE; 

/* Values for PP (assumes Ks=0, Kp=1) */
#define PP_RWXX	0	/* Supervisor read/write, User none */
#define PP_RWRX 1	/* Supervisor read/write, User read */
#define PP_RWRW 2	/* Supervisor read/write, User read/write */
#define PP_RXRX 3	/* Supervisor read,       User read */

/* Segment Register */
typedef struct _SEGREG
   {
      unsigned long t:1;	/* Normal or I/O  type */
      unsigned long ks:1;	/* Supervisor 'key' (normally 0) */
      unsigned long kp:1;	/* User 'key' (normally 1) */
      unsigned long n:1;	/* No-execute */
      unsigned long :4;		/* Unused */
      unsigned long vsid:24;	/* Virtual Segment Identifier */
   } SEGREG;

/* Block Address Translation (BAT) Registers */
typedef struct _BATU		/* Upper part of BAT */
   {
      unsigned long bepi:15;	/* Effective page index (virtual address) */
      unsigned long :4;		/* Unused */
      unsigned long bl:11;	/* Block size mask */
      unsigned long vs:1;	/* Supervisor valid */
      unsigned long vp:1;	/* User valid */
   } BATU;   

typedef struct _BATL		/* Lower part of BAT */
   {
      unsigned long brpn:15;	/* Real page index (physical address) */
      unsigned long :10;	/* Unused */
      unsigned long w:1;	/* Write-thru cache */
      unsigned long i:1;	/* Cache inhibit */
      unsigned long m:1;	/* Memory coherence */
      unsigned long g:1;	/* Guarded (MBZ) */
      unsigned long :1;		/* Unused */
      unsigned long pp:2;	/* Page access protections */
   } BATL;

typedef struct _BAT
   {
      BATU batu;		/* Upper register */
      BATL batl;		/* Lower register */
   } BAT;

/* Block size masks */
#define BL_128K	0x000
#define BL_256K 0x001
#define BL_512K 0x003
#define BL_1M   0x007
#define BL_2M   0x00F
#define BL_4M   0x01F
#define BL_8M   0x03F
#define BL_16M  0x07F
#define BL_32M  0x0FF
#define BL_64M  0x1FF
#define BL_128M 0x3FF
#define BL_256M 0x7FF

/* BAT Access Protection */
#define BPP_XX	0x00		/* No access */
#define BPP_RX	0x01		/* Read only */
#define BPP_RW	0x02		/* Read/write */

/* Used to set up SDR1 register */
#define HASH_TABLE_SIZE_64K	0x00010000
#define HASH_TABLE_SIZE_128K	0x00020000
#define HASH_TABLE_SIZE_256K	0x00040000
#define HASH_TABLE_SIZE_512K	0x00080000
#define HASH_TABLE_SIZE_1M	0x00100000
#define HASH_TABLE_SIZE_2M	0x00200000
#define HASH_TABLE_SIZE_4M	0x00400000
#define HASH_TABLE_MASK_64K	0x000   
#define HASH_TABLE_MASK_128K	0x001   
#define HASH_TABLE_MASK_256K	0x003   
#define HASH_TABLE_MASK_512K	0x007
#define HASH_TABLE_MASK_1M	0x00F   
#define HASH_TABLE_MASK_2M	0x01F   
#define HASH_TABLE_MASK_4M	0x03F   

/* invalidate a TLB entry */
extern inline void _tlbie(unsigned long va)
{
	asm volatile ("tlbie %0" : : "r"(va));
}

extern void _tlbia(void);		/* invalidate all TLB entries */
#endif

#if 1
/* Number of TLB entries, and how many to reserve */
#define NTLB		64
extern int NTLB_WIRED;

#define TLBHI_EPN	0xfffffc00
#define TLBHI_PGSZ	0x380
#define		TLBHI_PGSZ_1K	0
#define		TLBHI_PGSZ_4K	(1 << 7)
#define		TLBHI_PGSZ_16K	(2 << 7)
#define		TLBHI_PGSZ_64K	(3 << 7)
#define		TLBHI_PGSZ_256K	(4 << 7)
#define		TLBHI_PGSZ_1M	(5 << 7)
#define		TLBHI_PGSZ_4M	(6 << 7)
#define		TLBHI_PGSZ_16M	(7 << 7)
#define TLBHI_VALID	0x40

#define TLBLO_RPN	0xfffffc00
#define TLBLO_PERM	0x300
#define 	TLBLO_EX	0x200
#define 	TLBLO_WR	0x100
#define TLBLO_ZSEL	0xf0
#define		TLBLO_Z_KERN	0
#define		TLBLO_Z_USER	0x10
#define TLBLO_ATTR	0xf
#define 	TLBLO_W		0x8
#define 	TLBLO_I		0x4
#define 	TLBLO_M		0x2
#define 	TLBLO_G		0x1

#endif

#endif
