#ifndef __ASM_PPC_PROCESSOR_H
#define __ASM_PPC_PROCESSOR_H

/*
 * PowerPC machine specifics
 */

#define KERNEL_STACK_SIZE (4096) /* usable stack -- not buffers at either end */
#define KERNEL_STACK_MASK (~(KERNEL_STACK_SIZE-1))

/* Bit encodings for Machine State Register (MSR) */
#define MSR_ILE		(1<<16)		/* Interrupt Little-Endian enable */
#define MSR_EE		(1<<15)		/* External Interrupt enable */
#define MSR_PR		(1<<14)		/* Supervisor/User privilege */
#define MSR_ME		(1<<12)		/* Machine Check enable */
#define MSR_IP		(1<<6)		/* Exception prefix 0x000/0xFFF */
#define MSR_IR		(1<<5)		/* Instruction MMU enable */
#define MSR_DR		(1<<4)		/* Data MMU enable */
#define MSR_LE		(1<<0)		/* Little-Endian enable */

#if 0
#define MSR_POW		(1<<18)		/* Enable Power Management */
#define MSR_TGPR	(1<<17)		/* TLB Update registers in use */
#define MSR_FP		(1<<13)		/* Floating Point enable */
#define MSR_FE0		(1<<11)		/* Floating Exception mode 0 */
#define MSR_SE		(1<<10)		/* Single Step */
#define MSR_BE		(1<<9)		/* Branch Trace */
#define MSR_FE1		(1<<8)		/* Floating Exception mode 1 */
#define MSR_RI		(1<<1)		/* Recoverable Exception */

#define MSR_KERNEL	MSR_ME|MSR_FE0|MSR_FE1|MSR_IR|MSR_DR|MSR_RI
#endif

#if 1
/* PPC 403 */
#define MSR_WAIT	(1<<18)		/* Enter Wait State */
#define MSR_CI		(1<<17)		/* Critical Exception Enable */
#define MSR_DE		(1<<9)		/* Debug Exception */
#define MSR_PE		(1<<3)		/* Protection Enable */
#define MSR_PX		(1<<2)		/* Protection Exclusive Mode */

#ifdef CONFIG_XMON
#define MSR_KERNEL	MSR_CI|MSR_DE|MSR_ME|MSR_IR|MSR_DR
#else
#define MSR_KERNEL	MSR_CI|MSR_ME|MSR_IR|MSR_DR
#endif
#endif

#ifdef CONFIG_XMON
#define MSR_USER	MSR_KERNEL|MSR_DE|MSR_PR|MSR_EE
#else
#define MSR_USER	MSR_KERNEL|MSR_PR|MSR_EE
#endif

#if 0
/* Bit encodings for Hardware Implementation Register (HID0)
   on PowerPC 603 processors. */
#define HID0_EMCP	(1<<31)		/* Enable Machine Check pin */
#define HID0_EBA	(1<<29)		/* Enable Bus Address Parity */
#define HID0_EBD	(1<<28)		/* Enable Bus Data Parity */
#define HID0_SBCLK	(1<<27)
#define HID0_EICE	(1<<26)
#define HID0_ECLK	(1<<25)
#define HID0_PAR	(1<<24)
#define HID0_DOZE	(1<<23)
#define HID0_NAP	(1<<22)
#define HID0_SLEEP	(1<<21)
#define HID0_DPM	(1<<20)
#define HID0_ICE	(1<<15)		/* Instruction Cache Enable */
#define HID0_DCE	(1<<14)		/* Data Cache Enable */
#define HID0_ILOCK	(1<<13)		/* Instruction Cache Lock */
#define HID0_DLOCK	(1<<12)		/* Data Cache Lock */
#define HID0_ICFI	(1<<11)		/* Instruction Cache Flash Invalidate */
#define HID0_DCI	(1<<10)		/* Data Cache Invalidate */
#define HID0_SIED	(1<<7)		/* Serial Instruction Execution [Disable] */
#define HID0_BHTE	(1<<2)		/* Branch History Table Enable */
#define HID0_BTCD	(1<<1)		/* Branch target cache disable */
#endif

/* fpscr settings */
#define FPSCR_FX        (1<<31)
#define FPSCR_FEX       (1<<30)

/* task flags that assembly code needs to see */
#define TASK_FLAG_SS	0x1		/* single-step mode */

#ifndef __ASSEMBLY__

void start_thread(struct pt_regs *regs, unsigned long nip, unsigned long sp);
struct user;
void dump_thread(struct pt_regs *regs, struct user *dump);

/*
 * Bus types
 */
#define EISA_bus 0
#define EISA_bus__is_a_macro /* for versions in ksyms.c */
#define MCA_bus 0
#define MCA_bus__is_a_macro /* for versions in ksyms.c */

/*
 * Write Protection works right in supervisor mode on the PowerPC
 */
#define wp_works_ok 1
#define wp_works_ok__is_a_macro /* for versions in ksyms.c */

/*
 * User space process size: 2GB. This is hardcoded into a few places,
 * so don't change it unless you know what you are doing.
 */
#define TASK_SIZE	(0x80000000UL)

/* This decides where the kernel will search for a free chunk of vm
 * space during mmap's.
 */
#define TASK_UNMAPPED_BASE     (TASK_SIZE / 8 * 3)

struct thread_struct {
   	unsigned long	ksp;		/* Kernel stack pointer */
	unsigned long	*pg_tables;	/* MMU information */
#if defined(CPU) && (CPU != 403)
	unsigned long	segs[16];	/* MMU Segment registers */
#endif
	unsigned long	last_pc;	/* PC when last entered system */
	unsigned long	user_stack;	/* [User] Stack when entered kernel */
	double		fpr[32];	/* Complete floating point set */
	unsigned long	fpscr_pad;	/* fpr ... fpscr must be contiguous */
	unsigned long	fpscr;		/* Floating point status */
	unsigned long	wchan;		/* Event task is sleeping on */
	struct pt_regs	*regs;		/* Pointer to saved register state */
	unsigned long	fs;		/* for get_fs() validation */
	unsigned long	flags;		/* (pad to 16-byte bdry) */
	unsigned long	bear;		/* Bus Error Address Register */
	unsigned long	besr;		/* Bus Error Syndrome Register */
	unsigned long   dear;       	/* Data Error Address Register */

	/*
	 * These are only used during TLB handling
	 * operations with interrupts off.
	 */
	unsigned long	lastpte;
	unsigned long	tmp1;
	unsigned long	tmp2;
	unsigned long	ccr;
};

/* Points to the thread_struct of the thread (if any) which
   currently owns the FPU. */
extern struct thread_struct *fpu_tss;

#ifndef CONFIG_TCD
#define INIT_TSS  { \
	sizeof(init_kernel_stack) + (long) &init_kernel_stack,\
	(long *)swapper_pg_dir, {0}, \
	0, 0, {0}, 0, 0, \
	0, 0, \
	KERNEL_DS, 0, 0, \
}
#else
#define INIT_TSS  { \
	sizeof(init_kernel_stack) + (long) &init_kernel_stack,\
	(long *)swapper_pg_dir, \
	0, 0, {0}, 0, 0, \
	0, 0, \
	KERNEL_DS, 0, 0, \
	0, \
}
#endif

#define INIT_MMAP { &init_mm, 0, 0x40000000, \
		      PAGE_SHARED, VM_READ | VM_WRITE | VM_EXEC }

/* Free all resources held by a thread. */
extern void release_thread(struct task_struct *);

/*
 * Return saved PC of a blocked thread. For now, this is the "user" PC
 */
static inline unsigned long thread_saved_pc(struct thread_struct *t)
{
	return (t->last_pc);
}

#define _PROC_Motorola 0
#define _PROC_IBM      1
#define _PROC_Be       2

extern int _Processor;

/* Allocation and freeing of basic task resources. */
#define alloc_task_struct()	kmalloc(sizeof(struct task_struct), GFP_KERNEL)
#define free_task_struct(p)	kfree(p)

#ifdef KERNEL_STACK_BUFFER
/* give a 1 page buffer below the stack */
#define alloc_kernel_stack()  \
          (memset((void *)__get_free_pages(GFP_KERNEL,1,0),0,KERNEL_STACK_SIZE+PAGE_SIZE)+PAGE_SIZE)
#define free_kernel_stack(page) free_pages((page)-PAGE_SIZE,1)
#else

#define alloc_kernel_stack(p)	__get_free_page(GFP_KERNEL)
#define free_kernel_stack(page)	free_page((page))
#endif

#endif /* ASSEMBLY*/

#endif
